﻿using System.Diagnostics;
using System.IO;
using System.Net;
using System.Net.Security;
using System.Net.Sockets;
using System.Security.Authentication;
using System.Security.Cryptography.X509Certificates;
using System.Threading;

// Adapted from https://msdn.microsoft.com/en-us/library/system.net.security.sslstream.aspx
namespace SecureEchoTcpServer
{
    public class SecureEchoServer
    {
        public const int DefaultPort = 7777;
        public const string ServerCertificateFilename = "c:/temp2/anbo2.cer";
        public int Port { get; }
        public IPAddress LocalAddress { get; }
        public X509Certificate ServerSertificate { get; }

        private volatile bool _listening = true;
        private TcpListener _server;

        public SecureEchoServer(int port = DefaultPort)
        {
            Port = port;
            LocalAddress = IPAddress.Loopback;
            ServerSertificate = X509Certificate.CreateFromCertFile(ServerCertificateFilename);       
        }

        public void Start()
        {
            _server = new TcpListener(LocalAddress, Port);
            _server.Start();
            Trace.WriteLine("Secure Echo TCP server started on " + LocalAddress + " port " + Port);
            while (_listening)
            {
                if (!_server.Pending()) //checking for pending connections 
                {
                    Thread.Sleep(50);
                    continue;
                }
                TcpClient client = _server.AcceptTcpClient();
                Trace.WriteLine("Client connected: " + client);
                ProcessClient(client);
            }
        }

        public void Stop()
        {
            _listening = false;
            _server.Stop();
        }

        private void ProcessClient(TcpClient client)
        {
            NetworkStream ordinaryStream = client.GetStream();
            bool leaveInnerStreamOpen = false;
            // Decorator/Wrapper design pattern
            SslStream sslStream = new SslStream(ordinaryStream, leaveInnerStreamOpen);
            try
            {
                Authenticate(sslStream);
                StreamReader reader = new StreamReader(sslStream);
                string request = reader.ReadLine();
                Trace.WriteLine("Server read request: " + request);
                StreamWriter writer = new StreamWriter(sslStream);
                writer.WriteLine(request);
                writer.Flush();
            }
            catch (AuthenticationException ex)
            {
                Trace.WriteLine("Server exception: " + ex.Message);
            }
            catch (IOException ex)
            {
                Trace.WriteLine("Server IO exception: " + ex.Message);
            }
            finally
            {
                sslStream.Close();
                client.Close();
            }
        }

        private void Authenticate(SslStream sslStream)
        {
            bool clientCertificateRequired = false;
            bool checkCertificateRevocation = true;
            SslProtocols enabledSslProtocols = SslProtocols.Tls;
            sslStream.AuthenticateAsServer(ServerSertificate,
                clientCertificateRequired,
                enabledSslProtocols,
                checkCertificateRevocation);
            //sslStream.AuthenticateAsServer(ServerSertificate);            
        }
    }
}